require 'tmpdir'
require 'yaml'
require 'velcro/concerns/flags'
require 'velcro/concerns/shell'

module Velcro
  # The main entry point for velcro
  class Strap
    include Velcro::Flags
    include Velcro::Shell

    attr_reader :rcfile, :host

    FLAGS = %w(apps dotfiles symlinks scaffold settings).freeze

    def initialize(repo, options = {})
      establish_inclusions(FLAGS, options)
      if options.key?(:host)
        @host = options[:host]
        fail ArgumentError if host.nil? || host == ''
      end

      @dryrun = options[:dryrun] || false
      fail ArgumentError if repo.nil? && included?(:dotfiles)
      Velcro.config.repo_url = repo if repo
    end

    # rubocop:disable Metrics/AbcSize
    def run
      install_dotfiles

      Symlinker.new(dryrun: dryrun?).symlink! if included?(:symlinks)

      install_scaffold(rcfile)
      install_apps(rcfile['apps'])
      apply_settings(rcfile['settings'])

      if host
        set_hostname unless host == Velcro.host

        puts "\nSetting up #{host_data['name']}:"
        install_scaffold(host_data)
        install_apps(host_data['apps'])
        apply_settings(host_data['settings'])
      end
    end
    # rubocop:enable Metrics/AbcSize

    private

    def install_apps(apps)
      return unless included?(:apps)
      return if (apps ||= []).empty?

      puts "\nInstalling apps:"
      apps.each{|app| installer.install(app) }
    end

    def apply_settings(settings)
      return unless included?(:settings)
      return if (settings ||= []).empty?

      puts "\nApplying settings:"
      settings.each{|setting| setter.apply(setting) }
    end

    def install_scaffold(config)
      return unless included?(:scaffold)
      scaffolder.install(config)
    end

    def set_hostname
      shell("sudo scutil --set LocalHostName #{host}")
      shell("sudo scutil --set ComputerName #{host}")
    end

    def install_dotfiles
      if included?(:dotfiles)
        tmpdir = clone_dotfiles
        @rcfile = load_rcfile(tmpdir)

        shell("mv #{tmpdir} #{Velcro.config.dotfiles}")
        puts "Dotfiles installed: #{Velcro.config.dotfiles}"
      else
        fail MissingDotfilesError unless Status::Dotfiles.installed?
        @rcfile = load_rcfile(Velcro.config.dotfiles)
      end
    end

    def clone_dotfiles
      tmpdir = Dir.mktmpdir
      shell("git clone -q #{Velcro.config.repo_url} #{tmpdir}")
      tmpdir
    end

    def load_rcfile(dotfiles_dir)
      file = YAML.load_file(File.join(dotfiles_dir, '.velcrorc'))
      Velcro.config.apply file['config']
      file
    end

    def installer
      @installer ||= Installer::Base.installer(dryrun: dryrun)
    end

    def setter
      @setter ||= Setter::Base.setter(dryrun: dryrun)
    end

    def scaffolder
      @scaffolder ||= Scaffold.new(dryrun: dryrun)
    end

    def host_data
      return @host_data if defined?(@host_data)
      @host_data = rcfile['hosts'].detect{|n| n['name'] == Velcro.hostname }
      puts "No host settings found for: #{Velcro.hostname}" if host_data.nil?

      @host_data
    end
  end
end
